import appdirs
import configparser
import os.path
from collections import OrderedDict

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

_dirs = appdirs.AppDirs('torefl', 'hl037')

_defaults = {
		'edit_cmd' : 'vim "$f"',
		'edit_same_win' : '1',
		'open_cmd' : 'okular "$f"',
		'open_same_win' : '0',
}

_initial = OrderedDict((
		('DEFAULT', _defaults),
		('Torefl', _defaults),
		))

conf = configparser.ConfigParser()

conf.read_dict(_initial)

def write():
	with open(os.path.join(_dirs.user_config_dir, 'torefl.ini'), 'w') as f:
		conf.write(f)

if not os.path.isdir(_dirs.user_config_dir):
	os.makedirs(_dirs.user_config_dir)

try:
	with open(os.path.join(_dirs.user_config_dir, 'torefl.ini'), 'r') as f:
		conf.read_file(f, 'torefl.ini')
except:
	write()


	






import bibtexparser
import copy

from torefl import Entry
from typing import Iterable

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

def _getBib(e:Entry):
	b = copy.deepcopy(e.bibtex)
	b['ID'] = e.name
	return b	

def export(f, entries:Iterable[Entry] ):
	db = bibtexparser.bibdatabase.BibDatabase()
	db.entries = [ _getBib(e) for e in entries if e.bibtex is not None ]
	bibtexparser.dump(db, f)







import os
import bibtexparser

from sortedcontainers import SortedListWithKey as slist, SortedSet as sset
from typing import List, Tuple
from functools import reduce

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

class Existing(RuntimeError):
	pass


class Store(object):
	"""
	
	"""
	def __init__(self):
		self.l = []
		self.s = []
		self.i = 0
	
	def insert(self, x):
		if self.l:
			i = self.l.pop(0)
			self.s[i] = x
			return i
		else:
			self.s.append(x)
			return len(self.s) - 1
	
	def release(self, ID):
		self.s[ID] = None
		self.l.insert(0, ID)
	
	def replace(self, ID, x):
		
		if self.s[ID] is None:
			self.l.remove(ID)
		self.s[ID] = x
	
	def capacity(self):
		return len(self.s)
	
	def __len__(self):
		return len(self.s) - len(self.l)

	def __getitem__(self, k):
		return self.s[k]

	def values(self):
		return (s for s in self.s if s is not None)

	def container(self):
		return self.s
	
	def __iter__(self):
		return iter(enumerate(self.s))

	def idMax(self):
		i = -1
		s = self.s
		m = -len(s) - 1
		while s[i] is None:
			i-=1
			if i is m:
				return None
		return len(s) + i
	
	def reserve(self, nb):
		s = self.s
		curS = len(s)
		if nb <= curS:
			return
		s.extend([None] * (nb-curS))
		self.l.extend(list(range(curS, nb)))
		
		

class StoreWithName(Store):
	def __init__(self):
		super().__init__()
		self.byName = {}
	
	def insert(self, x):
		self.byName[x.name] = x
		return super().insert(x)

	def replace(self, ID, x):
		old = self[ID]
		if old is None:
			self.byName[x.name] = x
		elif old.name != x.name:
			del self.byName[old.name]
			self.byName[x.name] = x
		return super().replace(ID, x)
	
	def release(self, ID):
		x = self[ID]
		del self.byName[x.name]
		return super().release(ID)
	
	def __len__(self):
		return len(self.byName)
	
	def fromName(self, n):
		return self.byName[n]

class Entry(object):
	def __init__(self, name : str, bibtex : dict, tags : List[str], priority : float, depends : None, url = '', comment=''):
		self.name = name
		self.ID = None
		self.bibtex = bibtex
		self.comment = comment
		self.url = url
		self.tags = tags
		self.priority = priority
		self.depends = depends
		self._location = None
		
	def path(self):
		if self._location:
			return self._location.path()
		return []
	
	def pathstr(self):
		if self._location:
			return '/' + self._location.pathstr()
		return ''
	
	def fullpathstr(self):
		if self._location:
			s = self._location.pathstr()
			if s == '':
				return '/' + self.name
			else:
				return '/' + s + '/' + self.name
		return ''
	
	def toDict(self):
		return {'ID' : self.ID, 'comment' : self.comment, 'tags': self.tags, 'priority' : self.priority, 'depends': self.depends, 'path' : self.path() }
	
	@staticmethod
	def fromDict(d):
		t = Entry(d['content'], d['tags'], d['priority'], d['depends'])
		t.ID = d['ID']
		return t


_priorityFn = lambda x: x.priority
_path_priorityFn = lambda x: (x.pathstr(), x.priority)



class EntryRepo(object):
	"""
	
	"""
	def __init__(self, name = '', parent = None):
		self.children = {}
		self.parent = parent
		self.entries = slist(key=_priorityFn)
		self.name = name
		self._pathChanged = True
		self._path = []
		self._pathstr = []
	
	def add(self, t):
		self.entries.add(t)
		t._location = self
	
	def rem(self, t):
		self.entries.remove(t)
	
	def clear(self):
		self.children = {}
		self.entries = slist(key=_priorityFn)
	
	def _updatePath(self):
		if self.parent is None:
			self._path = []
		else:
			self._path = self.parent.path() + [self.name]
		self._pathstr = '/'.join(self._path)
		self._pathChanged = False
	
	def path(self):
		if self._pathChanged :
			self._updatePath()
		return self._path
	
	def pathstr(self):
		if self._pathChanged :
			self._updatePath()
		return self._pathstr
	
	def __getitem__(self, k):
		if not isinstance(k, str):
			return reduce(lambda x, y: x[y], k, self)
		if k == '':
			return self
		v = self.children.get(k)
		if v is None:
			v = EntryRepo(k, self)
			self.children[k] = v
		return v
	
	@staticmethod
	def p_br(p, l):
		return l.bisect_key_right(p)
	
	@staticmethod
	def p_bl(p, l):
		return l.bisect_key_left(p)
	
	@staticmethod
	def noMin(l):
		return None
	
	@staticmethod
	def noMax(l):
		return None
	
	def filter_legacy(self, tags, minFn, maxFn, pattern):
		entries = self.entries
		match = pattern.fullmatch if pattern else lambda x: True
		yield from ( t for t in entries[minFn(entries):maxFn(entries)] if all(tag in t.tags for tag in tags) and match(t.pathstr()) )
		for v in self.children.values():
			yield from v.filter_legacy(tags, minFn, maxFn, pattern)



class TagRepo(object):
	"""
	
	"""
	def __init__(self):
		self.tags = {}
	
	def __getitem__(self, k):
		v = self.tags.get(k)
		if v is None:
			v = slist(key=path_priorityFn)
			self.tags[k] = v
		return v
	
	def rem(self, entry):
		for t in entry.tags:
			self.tags[t].remove(entry)
	
	def add(self, entry):
		for t in entry.tags:
			tl = self.tags.get(t)
			if tl is None:
				self.tags[t] = slist((entry,), key=_priorityFn)
			else:
				tl.add(entry)
	
	def clear(self):
		self.tags = {}


class Torefl(object):
	"""
	
	"""
	def __init__(self, location=os.getcwd()):
		self.location = location
		self.store = StoreWithName()
		self.root = EntryRepo()
		self.tags = TagRepo()
		self.entries = slist(key=lambda x:(x.location(), x.priority))
	
	def reserve(self, nb):
		self.store.reserve(nb)
	
	def add(self, entry, path):
		entry.ID = self.store.insert(entry)
		self.root[path].add(entry)
		self.tags.add(entry)
	
	def place(self, ID, entry, path):
		if self.store[ID] is not None:
			raise ValueError('ID '+ID+' already taken')
		entry.ID = ID
		self.store.replace(ID, entry)
		self.root[path].add(entry)
		self.tags.add(entry)
	
	def update(self, entry, newEntry):
		newEntry.ID = entry.ID
		self.store.replace(entry.ID, newEntry)
		rep = self.root[entry.path()]
		rep.rem(entry)
		rep.add(newEntry)
		self.tags.rem(entry)
		self.tags.add(newEntry)





import re
import fnmatch
import operator

from typing import List, Tuple
from torefl.parser import *
from torefl.core import Entry

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

class BibFilter(object):
	__slots__ = ('checkList')
	def __init__(self, checkList : List[Tuple[str, str]]):
		self.checkList = checkList
	
	def __call__(self, entry : Entry):
		if entry.bibtex is None:
			return False
		b = entry.bibtex
		try:
			if not all((v not in b[k[1:]].lower()) if k[0] == '!' else (v in b[k].lower()) for k, v in self.checkList):
				return False
		except KeyError:
			return False
		else:
			return True

class BibFilterHandler(object):
	def __init__(self):
		self.tokens = []
	
	def __call__(self, tok : str):
		self.tokens.append(tok.lower())
	
	def get(self):
		l = self.tokens
		return BibFilter([(l[i], l[i+1]) for i in range(0, len(l), 2)])

class CommentFilter(object):
	__slots__ = ('checks')
	def __init__(self, checks : List[str]):
		self.checks = checks
	
	def __call__(self, entry : Entry):
		c = entry.comment.lower()
		return all(s in c for s in self.checks)

class PathFilter(object):
	__slots__ = ('re',)
	def __init__(self, path):
		self.re = re.compile(fnmatch.translate(path.lower()))
	def __call__(self, entry: Entry):
		return bool(self.re.fullmatch(entry.fullpathstr().lower()))

class TagFilter(object):
	__slots__ = ('tag',)
	def __init__(self, tag):
		self.tag = formatTag(tag)
	def __call__(self, entry : Entry):
		return self.tag in entry.tags

class PriorityFilter(object):
	__slots__ = ('fun', 'p')
	def __init__(self, p_match):
		c = p_match['cmp']
		if   c == '>=':
			self.fun = operator.ge
		elif c == '>':
			self.fun = operator.gt
		elif c == '<=':
			self.fun = operator.le
		elif c == '<':
			self.fun = operator.lt
		else:
			self.fun = operator.eq
		self.p = float(p_match['p'])
	def __call__(self, entry: Entry):
		return self.fun(entry.priority, self.p)

class AndFilter(object):
	__slots__ = ('filters')
	def __init__(self, filters):
		self.filters = filters
	def __call__(self, entry : Entry):
		return all(f(entry) for f in self.filters)

class NotFilter(object):
	__slots__ = ('filter')
	def __init__(self, filter):
		self.filter = filter
	def __call__(self, entry : Entry):
		return not self.filter(entry)

identity = lambda x : x
not_op = lambda x : NotFilter(x)
	

def LegacyFilter(tokens):
	filters = []
	for t in tokens:
		if t[0] == '!':
			f = not_op
			t = t[1:]
		else:
			f = identity
		m = path_f_re.fullmatch(t)
		if m:
			filters.append(f(PathFilter(m[0])))
			continue
		m = TagCollector.regex.fullmatch(t)
		if m:
			filters.append(f(TagFilter(m[1])))
			continue
		m = priority_f_re.fullmatch(t)
		if m:
			filters.append(f(PriorityFilter(m)))
			continue
	return AndFilter(filters)

class GenericFilterHandler(object):
	__slots__ = ('tokens','Filter','lower')
	def __init__(self, Filter, lower = True):
		self.tokens = []
		self.Filter = Filter
		self.lower = lower
		
	def __call__(self, tok : str):
		self.tokens.append(tok.lower() if self.lower else tok)
	
	def get(self):
		return self.Filter(self.tokens)


filterHandlers = {
		'bib' : BibFilterHandler,
		'com' : lambda : GenericFilterHandler(CommentFilter),
		'leg' : lambda : GenericFilterHandler(LegacyFilter, False),
}


def getFilterFunctions(args):
	h = filterHandlers['leg']()
	r = []
	for a in args:
		if a[0] == '-':
			r.append(h.get())
			h = filterHandlers[a[1:]]()
		else:
			h(a)
	r.append(h.get())
	return r





import unittest
import torefl.parser as parsers
import io

class TestBase(unittest.TestCase):
	"""
	Test reading torefl bases
	"""
	def setUp(self):
		buf = """_torefl000
		30
		10:Test1:A Title
		11:Test2:
		12:Test3:Another Title
		"""
		self.f = io.StringIO(buf)
	
	def test_ParseBase(self):
		d1, d2, n = parsers.parseBase(self.f)
		self.assertDictEqual(d1, {'Test1':10, 'Test2':11, 'Test3':12})
		self.assertDictEqual(d2, {'A Title':10, 'Another Title':12})
		self.assertEqual(n, 30)
	




import re
import bibtexparser
from torefl import Entry

from colorama import Fore as _F, Back as _B, Style as _S, init as _colorama_init

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"


def iterStrip(l):
	return iter(s.strip() for s in l)

def cap(s):
	return s[0].upper() + s[1:]

def formatTag(tag):
	l = tag.split('_')
	return ''.join(cap(s) for s in l)


class Collector(object):
	"""
	
	"""
	def __init__(self):
		self.l = []
	
	def __call__(self, match):
		self.l.append(self.parse(match))
		return ''

	def sub(self, s):
		return self.regex.sub(self, s)


class TagCollector(Collector):
	regex = re.compile(r'(?:^|\s)\#(\w+)')
	def parse(self, m):
		return formatTag(m[1])

class PriorityCollector(Collector):
	regex = re.compile(r'(?:^|\s)@((\d+\.\d+)|(\d+))')
	def parse(self, m):
		return float(m[1])

id_re = re.compile(r'(?:^|\s)&(\d+)')
ide_re = re.compile(r'(?:^|\s)&(\w+)')
path_re = re.compile(r'(/\w+)+|/')
path_f_re = re.compile(r'(/(\w|\*)+)+|/')
priority_f_re = re.compile(r'(?:^|\s)@(?P<cmp>(<=|>=|<|>)?)(?P<p>(\d+\.\d+)|(\d+))')

def parseComment(comment):
	tc = TagCollector()
	pc = PriorityCollector()
	comment = "".join( pc.sub(tc.sub(s)) for s in comment ).strip()
	return comment, tc.l, pc.l[0] if pc.l else float(0)

def parseEntry(f, name, pdf_path):
	lines = iter(f)
	comment = []
	bibtex = []
	for li in lines:
		if li.startswith('****************'):
			break
		comment.append(li)
	for li in lines:
		bibtex.append(li)
	if not bibtex: #if there is no separator, then we consider the file as a bibtex...
		bibtex, comment = comment, bibtex
	return createEntry(name, pdf_path, comment, bibtex)

def createEntry(name, pdf_path, comment, bibtex):
	bib = ''.join(bibtex)
	#Todo Catch exceptions
	try:
		db = bibtexparser.loads(bib)
		if db.entries:
			parsedBib = db.entries[0]
		else:
			print(_F.RED, 'No citation for ', name, _F.RESET, sep='')
			parsedBib = None
	except Exception as e:
		print('EX :', e.__class__.__name__)
		print(e)
		parsedBib = None
	comment, tag_list, priority = parseComment(comment)
	return Entry(name, parsedBib, tag_list, priority, None, url=pdf_path, comment=comment)

class UnsupportedVersion(RuntimeError):
	pass

base000_line_re = re.compile(r'(?P<ID>\d+):(?P<name>.+?):(?P<title>.*)')
base000_register_re = re.compile(r'(?P<ID>[^:]+?):(?P<sel>(\d+(\s+\d+)*)?)')

def parseBase000(it):
	nb = int(next(it))
	e = []
	for l in it:
		l = l.strip()
		if l.startswith('****************'):
			break
		t = base000_line_re.fullmatch(l)
		if t:
			e.append(t)
	r = []
	for l in it:
		l = l.strip()
		t = base000_register_re.fullmatch(l)
		if t:
			r.append(t)
	return	(
						{ m.group('name') : int(m.group('ID')) for m in e if m },
						{ m.group('title') : int(m.group('ID')) for m in e if m and m.group('title') != '' },
						nb,
						( (m.group('ID'), map(int, m.group('sel').split())) for m in r ),
					)

def parseBase(f):
	it = iterStrip(f)
	v = next(it)
	if v == '_torefl000':
		return parseBase000(it)
	else:
		raise UnsupportedVersion(v)
		





import operator
import sys
import cmd
import os
import sys
from functools import partial
from itertools import chain
import json
import bibtexparser
import shlex
import traceback
import subprocess
from typing import List, Tuple

from colorama import Fore as _F, Back as _B, Style as _S, init as _colorama_init
from torefl.formatters import ASCIIFormatter as _ASCIIFormatter
from torefl import Torefl, Entry
from torefl.filter import getFilterFunctions
from torefl.exporters import BibtexExporter
from torefl.config import conf, write as writeConf
from torefl.parser import parseEntry, parseBase

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

class _listHack(list):
	"""
	
	"""
	def __init__(self, g):
		self.g = g
	
	def __iter__(self):
		return iter(self.g)
	
	def __len__(self):
		return len(self.g)

	def __getitem__(self, k):
		return self.g[k]

_torefl = Torefl()

_store = _torefl.store
_root = _torefl.root
_tags = _torefl.tags
_entries = _torefl.entries


def _updateGlobal():
	global _store
	global _root
	global _tags
	global _entries
	_store = _torefl.store
	_root = _torefl.root
	_tags = _torefl.tags
	_entries = _torefl.entries

_updateGlobal()

def _GetOrInsert(defaultClass):
	def f(d, key):
		try:
			return d[key]
		except KeyError:
			r = defaultClass()
			d[key] = r
			return r
	return f

_getOrInsertSet = _GetOrInsert(set)

def _getById(torefl:Torefl, id) -> Entry:
	try:
		return torefl.store[int(id)]
	except:
		return torefl.store.fromName(id)

def _getPath(torefl:Torefl, e:Entry, ext:str):
	return os.path.join(torefl.location, e.pathstr()[1:], e.name + ext)

def _exec(line:str, same_win:bool):
	argv = shlex.split(line)
	if same_win:
		subprocess.run(argv)
	else:
		subprocess.Popen(argv, stdout = subprocess.DEVNULL)

class Shell(cmd.Cmd):
	intro = 'torefl shell'
	prompt = 'torefl / >'

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.selection = set()
		self.registers = {}
	
	@staticmethod
	def parse(arg):
		res = {}
		i = 0
		m = path_re.match(arg, i)
		if m:
			res['path'] = m[0][1:].split('/')
			arg = arg[m.end():].strip()
		else:
			res['path'] = []
			
		l = []
		m_tag, arg = getAll(TagCollector.regex, arg)
		res['tags'] = [ formatTag(m[1]) for m in m_tag ]
		
		m = PriorityCollector.regex.search(arg)
		if m:
			res['priority'] = float(m[1])
			arg = (arg[:m.start()] + arg[m.end():]).strip()
		else:
			res['priority'] = 0.
		res['_'] = arg
		return res
		
	def Rem(self, arg):
		for ID in map(int, arg.split()):
			try:
				t = _store[ID]
			except:
				print(_F.RED, 'ID %s Not found' % ID, _S.RESET_ALL, sep='')
			else:
				r = _root[t.path()[1:]]
				r.rem(t)
				_tags.rem(t)
				_store.release(ID)
				print(_F.BLUE, 'Entry removed : \n', str(t), sep='')
	
	def LegacyFilter(self, arg):
		m_path, arg = getAll(path_f_re, arg)
		if m_path:
			paths = [m[0] for m in m_path]
		else:
			paths = ['/']
		
		m_tag, arg = getAll(TagCollector.regex, arg)
		_tags = [ formatTag(m[1]) for m in m_tag ]
			
		m_priority, arg = getAll(priority_f_re, arg)
		minFn = EntryRepo.noMin
		maxFn = EntryRepo.noMax
		for m in m_priority:
			p = float(m['p'])
			c = m['cmp']
			if   c == '>=':
				minFn = partial(EntryRepo.p_bl, p)
			elif c == '>':
				minFn = partial(EntryRepo.p_br, p)
			elif c == '<=':
				maxFn = partial(EntryRepo.p_br, p)
			elif c == '<':
				maxFn = partial(EntryRepo.p_bl, p)
			else:
				minFn = partial(EntryRepo.p_bl, p)
				maxFn = partial(EntryRepo.p_br, p)
		return sset( chain.from_iterable( repo.filter_legacy(_tags, minFn, maxFn, pattern) for repo, pattern in (getRepoPath(x) for x in paths) ), key = lambda x: (x.pathstr(), x.priority, x.ID) )

	def Filter(self, args):
		lf = getFilterFunctions(args[1:])
		return [ e for e in _torefl.store.values() if all(f(e) for f in lf) ]
		

	def load(self, st):
		_tags.clear()
		_root.clear()
		o = json.load(st)
		_store.i = o['maxSize']
		print('maxSize : ', o['maxSize'])
		s = [None] * _store.i
		_store.s = s
		for t in o['entries']:
			print('ID : ', repr(t['ID']))
			ta = Entry.fromDict(t)
			s[t['ID']] = ta
			_tags.add(ta)
			_root[t['path']].add(ta)
		print(s)
		_store.l = [i for i, v in _store]
			
	def open(self, path):
		with open(path, 'r', encoding='utf-8') as f:
			self.load(f)
	
	def Load(self, arg):
		if arg:
			self.path = arg
		self.open(self.path, encoding='utf-8')
	
	def save(self, path):
		with open(path, 'w', encoding='utf-8') as st:
			_writeTorefl(st, _torefl, self.registers)
		

	def Save(self, arg):
		self.save(self.path if len(arg) == 0 else arg[1])
	
	def Quit(self, arg):
		if arg == '!':
			return True
		if arg:
			self.path = arg
		self.save(self.path)
		return True
	
	def List(self, arg):
		l = self.Filter(arg) 
		self.selection = set(l)
		for t in l :
			print(_ASCIIFormatter.formatEntry(t))
			
	def Sel(self, arg):
		if arg[0] in ('&=', '^=', '|=', '-=', '='):
			c = [ None, arg[0], arg[1] if len(arg) is 2 else None ]
		else:
			c = [ arg[0], arg[1], arg[2] if len(arg) is 3 else None ]
		if c[0] is None:
			s0 = self.selection
		else:
			s0 = _getOrInsertSet(self.registers, c[0])
		if c[2] is None:
			s1 = self.selection
		elif c[2] == '0':
			s1 = set()
		else:
			s1 = _getOrInsertSet(self.registers, c[2])
		if c[1] == '=':
			s0.clear()
			s0 |= s1
		else:
			exec('s0 ' + c[1] + 's1')
		
	def ListSel(self, arg):
		if len(arg) is 0:
			s = self.selection
		else:
			s = self.registers[arg[0]]
		for t in s:
			print(_ASCIIFormatter.formatEntry(t))
	
	def Export(self, arg):
		if arg[0] == 'bib':
			with open(arg[1], 'w', encoding='utf-8') as f:
				BibtexExporter.export(f, self.registers[arg[2]] if len(arg) > 2 else self.selection)
		else:
			print('Unknown exporter', arg[0])
	
	def Edit(self, arg):
		e = _getById(_torefl, arg[0])
		f = _getPath(_torefl, e, EXT)
		c = conf['Torefl'].get('edit_cmd')
		c = c.replace('$f', f)
		same_win = conf['Torefl'].getboolean('edit_same_win')
		_exec(c, same_win)
		with open(f, 'r') as f:
			n = parseEntry(f, e.name, e.url)
		_torefl.update(e, n)
		
	def Open(self, arg):
		e = _getById(_torefl, arg[0])
		f = _getPath(_torefl, e, '.pdf')
		c = conf['Torefl'].get('open_cmd')
		c = c.replace('$f', f)
		same_win = conf['Torefl'].getboolean('open_same_win')
		_exec(c, same_win)
	
	def default(self, line):
		argv = shlex.split(line)
		if argv[0] in ('l', 'list'):
			return self.List(argv)
		if argv[0] in ('s', 'sel', 'selection'):
			return self.Sel(argv[1:])
		if argv[0] in ('ls',):
			return self.ListSel(argv[1:])
		if argv[0] in ('e', 'export'):
			return self.Export(argv[1:])
		if argv[0] in ('ed', 'edit'):
			return self.Edit(argv[1:])
		if argv[0] in ('o', 'open'):
			return self.Open(argv[1:])
		
	def emptyline(self):
		pass

	def onecmd(self, s):
		try:
			return super().onecmd(s)
		except:
			traceback.print_exc()
	
	
	#do_a = Add
	#do_add = Add
	
	#do_r = Rem
	#do_rem = Rem
	#do_remove = Rem
	
	do_save = Save
	do_sa = Save
	
	#do_o = Load
	#do_open = Load
	
	do_q = Quit
	do_quit = Quit



EXT = '.torefl'
EXT_BASE = '.toreflbase'

def _writeTorefl(st, _torefl, regs):
	store = _torefl.store
	st.write('_torefl000\n')
	st.write(str(len(store))+ '\n')
	st.write(' '.join(str(i) for i in store.l ))
	for v in store.values() :
		if v is not None:
			st.write(str(v.ID) + ':' + v.name + ':' + (v.bibtex.get('title', '') if v.bibtex else '') + '\n')
	st.write('****************\n')
	for k, v in regs.items():
		st.write(k + ':' + ' '.join(str(e.ID) for e in v) + '\n')

def _writeCache(st, torefl):
	pass


def walkTorefl(p):
	ext_len = len(EXT)
	l = [p]
	current = []
	for _current, dirs, files in os.walk(p):
		rel = os.path.relpath(_current, p)
		if rel == '.':
			current = []
		else:
			current = rel.split(os.sep)
		for f in files:
			if f.endswith('.torefl'):
				name = f[:-len(EXT)]
				pdf_path = current+[name+'.pdf']
				if not os.path.exists(os.sep.join(pdf_path)):
					pdf_path = None
				with open(os.sep.join(current + [f]), encoding='utf-8') as bib:
					e = parseEntry(bib, name, pdf_path)
					yield e, current
		l.append(None)


def _create(args):
	global _torefl
	_torefl = Torefl()
	_updateGlobal()
	for e, path in walkTorefl(os.getcwd()):
		_torefl.add(e, path)
	o = args.file + EXT_BASE
	with open(o, 'w', encoding='utf-8') as of:
		_writeTorefl(of, _torefl, {})
	#with open('.'+o+'.cache', 'w', encoding='utf-8') as of:
	#	_writeCache(of, _torefl)
	return o, []

def _update(args):
	global _torefl
	_torefl = Torefl()
	_updateGlobal()
	o = args.file + EXT_BASE
	with open(o, 'r', encoding='utf-8') as f:
		d1, d2, n, r = parseBase(f)
	_torefl.reserve(n)
	l = []
	for e, path in walkTorefl(os.getcwd()):
		try:
			ID = d2[e.bibtex['title']]
		except:
			ID = d1.get(e.name, None)
		if ID is not None:
			_torefl.place(ID, e, path)
		else:
			l.append((e, path))
	for e, path in l:
		_torefl.add(e, path)
	r = { k : { _store[ID] for ID in v if _store[ID] is not None } for k, v in r }
	with open(o, 'w', encoding='utf-8') as of:
		_writeTorefl(of, _torefl, r)
	#with open('.'+o+'.cache', 'w', encoding='utf-8') as of:
	#	_writeCache(of, _torefl)
	return o, r

def _default(args):
	o = args.file + EXT_BASE
	if os.path.isfile(o):
		return _update(args)
	else:
		return _create(args)

	




def getAll(r, arg):
	res = []
	m = r.search(arg)
	while m:
		res.append(m)
		arg = arg[:m.start()].strip() + arg[m.end():]
		m = r.search(arg)
	return res, arg

def getRepoPath(x):
	i = x.find('*')
	if i == -1:
		pre = x.split('/')
		del pre[0]
		return root[pre], None
	else:
		pre = x[:i].split('/')
		del pre[0]
		del pre[-1]
		return root[pre], re.compile(fnmatch.translate(x))




import argparse

from torefl.core import *
import torefl.shell as shell

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

def main():
	parser = argparse.ArgumentParser()
	parser.add_argument('-q', action='store_true', help='Write cache and quit')
	parser.set_defaults(func=shell._default, file='biblio')
	sub = parser.add_subparsers(title='Action', description='Valid sub commands :')
	p_create = sub.add_parser('create', aliases=['c'], help='Create a bib tree')
	p_create.add_argument('file', action='store', nargs='?', default='biblio', help='Name of the database file to create (without extension)')
	p_create.set_defaults(func=shell._create)
	p_update = sub.add_parser('update', aliases=['u'], help='Update a bib tree')
	p_update.add_argument('file', action='store', nargs='?', default='biblio', help='Name of the database to use/update')
	p_update.set_defaults(Action=p_update)
	args = parser.parse_args()
	path, r = args.func(args)
	if not args.q:
		sh = shell.Shell()
		sh.path = path
		sh.registers = r
		sh.cmdloop()





import unittest
from torefl import Store, StoreWithName
import io

class TestStore(unittest.TestCase):
	"""
	Test reading torefl bases
	"""
	def setUp(self):
		self.store = Store()
	
	def get(self, val):
		return val
	
	def prepareReserve(self):
		self.store.reserve(20)
		
	
	def test_0_reserve(self):
		self.prepareReserve()
		store = self.store
		self.assertEqual(store.capacity(), 20)
		self.assertListEqual(list(store.values()), [])
	
	def test_1_insert(self):
		store = self.store
		idA = self.idA = store.insert(self.get('A'))
		idB = self.idB = store.insert(self.get('B'))
		self.assertNotEqual(idA, idB)
		self.assertEqual(store[idA], self.get('A'))
		self.assertEqual(store[idB], self.get('B'))
		self.assertEquals(store.capacity(), 2)
		self.assertEquals(len(store), 2)
		self.assertEquals(store.idMax(), 1)
		l1 = list(store.values())
		l2 = [self.get('A'), self.get('B')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)
	
	def test_1_reserve_insert_less(self):
		store = self.store
		store.reserve(20)
		idA = self.idA = store.insert(self.get('A'))
		idB = self.idB = store.insert(self.get('B'))
		self.assertNotEqual(idA, idB)
		self.assertEqual(store[idA], self.get('A'))
		self.assertEqual(store[idB], self.get('B'))
		self.assertEquals(store.capacity(), 20)
		self.assertEquals(len(store), 2)
		self.assertEquals(store.idMax(), 1)
		l1 = list(store.values())
		l2 = [self.get('A'), self.get('B')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)
		
	def test_1_reserve_insert_more(self):
		store = self.store
		store.reserve(2)
		idA = self.idA = store.insert(self.get('A'))
		idB = self.idB = store.insert(self.get('B'))
		idC = self.idC = store.insert(self.get('C'))
		idD = self.idD = store.insert(self.get('D'))
		self.assertEqual(len({idA, idB, idC, idD}), 4)
		self.assertEqual(store[idA], self.get('A'))
		self.assertEqual(store[idB], self.get('B'))
		self.assertEqual(store[idC], self.get('C'))
		self.assertEqual(store[idD], self.get('D'))
		self.assertEquals(store.capacity(), 4)
		self.assertEquals(len(store), 4)
		self.assertEquals(store.idMax(), 3)
		l1 = list(store.values())
		l2 = [self.get('A'), self.get('B'), self.get('C'), self.get('D')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)
	
	def test_2_release(self):
		self.test_1_insert()
		store = self.store
		idA = self.idA
		idB = self.idB
		store.release(idA)
		self.assertEqual(list(store.values()), [self.get('B')])
		self.assertEquals(store.capacity(), 2)
		self.assertEqual(len(store), 1)
		self.assertEquals(store.idMax(), 1)
		idA2 = store.insert(self.get('A2'))
		self.assertEqual(idA2, idA, 'Test if it reuses the same id')
		l1 = list(store.values())
		l2 = [self.get('A2'), self.get('B')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)
		self.assertEquals(store.capacity(), 2)
		self.assertEqual(len(store), 2)
		self.assertEquals(store.idMax(), 1)
		
	def test_3_replace(self):
		self.test_1_insert()
		store = self.store
		idA = self.idA
		idB = self.idB
		idC = idB
		store.replace(idC, self.get('C'))
		self.assertEqual(store[idA], self.get('A'))
		self.assertEqual(store[idC], self.get('C'))
		self.assertEquals(len(store), 2)
		l1 = list(store.values())
		l2 = [self.get('A'), self.get('C')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)
		
	def test_3_replaceNone(self):
		self.test_1_reserve_insert_less()
		store = self.store
		idA = self.idA
		idB = self.idB
		idC = idA + idB + 1
		store.replace(idC, self.get('C'))
		self.assertEqual(store[idA], self.get('A'))
		self.assertEqual(store[idB], self.get('B'))
		self.assertEqual(store[idC], self.get('C'))
		self.assertEquals(store.capacity(), 20)
		self.assertEqual(len(store), 3)
		self.assertEquals(store.idMax(), idC)
		l1 = list(store.values())
		l2 = [self.get('A'), self.get('B'), self.get('C')]
		l1.sort()
		l2.sort()
		self.assertListEqual(l1, l2)

class _Obj(object):
	"""
	
	"""
	def __init__(self, val):
		self.val = val
	
	def __eq__(self, obj):
		try:
			return self.val == obj.val
		except:
			return self.val == obj

class TestStoreWithName(TestStore):
	"""
	
	"""
	def setUp(self):
		self.store = StoreWithName()
	
	def get(self, val):
		return _Obj(val)

	




from colorama import Fore as _F, Back as _B, Style as _S, init as _colorama_init
from torefl.core import Entry

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

def formatEntry(e:Entry):
	R = _S.RESET_ALL
	ID = R + _B.BLUE
	P = R + _F.LIGHTGREEN_EX
	TAG = R + _F.RED
	NAME = R + _B.RED
	Bd = R + _F.CYAN
	PATH = R + _F.YELLOW
	TITLE = R + _F.LIGHTMAGENTA_EX
	AUTHORS = R + _F.MAGENTA
	prefix = Bd + '|  ' + R
	comment = ( s + '\n' + prefix for s in e.comment.split('\n') ) if e.comment != '' else ()
	
	return ''.join( (
		Bd, '--------------------------------------------------------------------------------', R, '\n',
		prefix, ID, 'ID : ', '{:>5}'.format(e.ID or ''), R, ' '*47, P, '{:>20}'.format(e.priority), R, '\n',
		prefix, NAME, e.name, R, '\n',
		prefix, PATH, e.pathstr(), R, '\n',
		*( (
			prefix, TITLE, e.bibtex.get('title', ''), R, '\n',
			prefix, AUTHORS, e.bibtex.get('author', ''), R, '\n',
			) if e.bibtex else (
			prefix, TITLE, '<No Bibtex>', R, '\n')
		),
		prefix, (R + ' ').join(''.join((TAG, '#', t)) for t in e.tags), '\n', 
		prefix, R, *comment , '\n',
		Bd, '--------------------------------------------------------------------------------', R, '\n',
		))




#!/usr/bin/env python
import sys

from setuptools import setup, find_packages

__copyright__ = "Copyright 2017, Léo Flaventin Hauchecorne"
__license__ = "GPLv3"

def read_file(name):
    """
    Read file content
    """
    f = open(name)
    try:
        return f.read()
    except IOError:
        print("could not read %r" % name)
        f.close()

LONG_DESC = read_file('README.rst')

EXTRAS = {}

if sys.version_info < (3,):
    EXTRAS['use_2to3'] = True

setup(
    name='torefl',
    version='1.0b3',
    description='Software to manage references',
    long_description=LONG_DESC,
    author='hl037',
    author_email='hl037.prog@gmail.com',
    url='https://github.com/hl037/torefl',
    license='GPL',
    packages=find_packages(),
    test_suite=None,
    include_package_data=True,
    zip_safe=False,
    install_requires = [
			'appdirs>=1.4.3', 
			'bibtexparser>=0.6.2',
			'colorama>=0.3.9',
			'sortedcontainers>=1.5.7',
			],
    extras_require=None,
    entry_points={
			'console_scripts' : [ 'torefl=torefl:main' ]
			},
		keywords=['torefl', 'bibtex', 'bibliography', 'reference', 'pdf', 'management'],
    classifiers=[
			'Development Status :: 4 - Beta',
			'Environment :: Console',
			'Intended Audience :: Developers',
			'Intended Audience :: Education',
			'Intended Audience :: Science/Research',
			'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
			'Natural Language :: English',
			'Operating System :: OS Independent',
			'Programming Language :: Python :: 3',
			'Topic :: Scientific/Engineering',
			'Topic :: Utilities',
			'Topic :: Other/Nonlisted Topic',
			],
    **EXTRAS
)
